from __future__ import print_function 

import contextlib 
import errno 
import fcntl 
import os 
import stat 
import subprocess 
import sys 
import time 
import warnings 
import zipfile 
import textwrap 
from distutils import log 

import macholib .util 
import pkg_resources 
from macholib .util import is_platform_file 
from modulegraph import zipio 
from modulegraph .find_modules import PY_SUFFIXES 

try :
    unicode 
except NameError :
    unicode =str 


try :
    import imp 
except ImportError :
    imp =None 




def os_path_islink (path ):
    warnings .warn ("Use zipio.islink instead of os_path_islink",DeprecationWarning )
    return zipio .islink (path )


def os_path_isdir (path ):
    warnings .warn ("Use zipio.isdir instead of os_path_isdir",DeprecationWarning )
    return zipio .islink (path )


def os_readlink (path ):
    warnings .warn ("Use zipio.readlink instead of os_readlink",DeprecationWarning )
    return zipio .islink (path )


def get_zip_data (path_to_zip ,path_in_zip ):
    warnings .warn ("Use zipio.open instead of get_zip_data",DeprecationWarning )
    zf =zipfile .ZipFile (path_to_zip )
    return zf .read (path_in_zip )


def path_to_zip (path ):
    """
    Returns (pathtozip, pathinzip). If path isn't in a zipfile pathtozip
    will be None
    """
    warnings .warn ("Don't use this function",DeprecationWarning )
    zf =zipfile .ZipFile (path_to_zip )
    orig_path =path 
    from distutils .errors import DistutilsFileError 

    if os .path .exists (path ):
        return (None ,path )

    else :
        rest =""
        while not os .path .exists (path ):
            path ,r =os .path .split (path )
            if not path :
                raise DistutilsFileError ("File doesn't exist: %s"%(orig_path ,))
            rest =os .path .join (r ,rest )

        if not os .path .isfile (path ):

            raise DistutilsFileError ("File doesn't exist: %s"%(orig_path ,))

        try :
            zf =zipfile .ZipFile (path )
            zf .close ()
        except zipfile .BadZipfile :
            raise DistutilsFileError ("File doesn't exist: %s"%(orig_path ,))

        if rest .endswith ("/"):
            rest =rest [:-1 ]

        return path ,rest 


def get_mtime (path ,mustExist =True ):
    """
    Get mtime of a path, even if it is inside a zipfile
    """
    warnings .warn ("Don't use this function",DeprecationWarning )
    try :
        return zipio .getmtime (path )

    except IOError :
        if not mustExist :
            return -1 

        raise 





gConverterTab ={}


def find_converter (source ):
    if not gConverterTab :
        for ep in pkg_resources .iter_entry_points ("py2app.converter"):
            function =ep .load ()
            if not hasattr (function ,"py2app_suffix"):
                print (
                "WARNING: %s does not have 'py2app_suffix' attribute"%(function )
                )
                continue 
            gConverterTab [function .py2app_suffix ]=function 

    basename ,suffix =os .path .splitext (source )
    try :
        return gConverterTab [suffix ]

    except KeyError :
        return None 


def copy_resource (source ,destination ,dry_run =0 ,symlink =0 ):
    """
    Copy a resource file into the application bundle
    """
    converter =find_converter (source )
    if converter is not None :
        converter (source ,destination ,dry_run =dry_run )
        return 

    if os .path .isdir (source ):
        if not dry_run :
            if not os .path .exists (destination ):
                os .mkdir (destination )
        for fn in zipio .listdir (source ):
            copy_resource (
            os .path .join (source ,fn ),
            os .path .join (destination ,fn ),
            dry_run =dry_run ,
            symlink =symlink ,
            )

    else :
        if symlink :
            if not dry_run :
                make_symlink (os .path .abspath (source ),destination )

        else :
            copy_file (source ,destination ,dry_run =dry_run ,preserve_mode =True )


def copy_file (
source ,
destination ,
preserve_mode =False ,
preserve_times =False ,
update =False ,
dry_run =0 ,
):
    while True :
        try :
            _copy_file (
            source ,destination ,preserve_mode ,preserve_times ,update ,dry_run 
            )
            return 
        except IOError as exc :
            if exc .errno !=errno .EAGAIN :
                raise 

            log .info (
            "copying file %s failed due to spurious EAGAIN, "
            "retrying in 2 seconds",
            source ,
            )
            time .sleep (2 )


def _copy_file (
source ,
destination ,
preserve_mode =False ,
preserve_times =False ,
update =False ,
dry_run =0 ,
):
    log .info ("copying file %s -> %s",source ,destination )
    with zipio .open (source ,"rb")as fp_in :
        if not dry_run :
            if os .path .isdir (destination ):
                destination =os .path .join (destination ,os .path .basename (source ))
            if os .path .exists (destination ):
                os .unlink (destination )

            with open (destination ,"wb")as fp_out :
                data =fp_in .read ()
                fp_out .write (data )

            if preserve_mode :
                mode =None 
                if hasattr (zipio ,"getmode"):
                    mode =zipio .getmode (source )

                elif os .path .isfile (source ):
                    mode =stat .S_IMODE (os .stat (source ).st_mode )

                if mode is not None :
                    os .chmod (destination ,mode )

            if preserve_times :
                mtime =zipio .getmtime (source )
                os .utime (destination ,(mtime ,mtime ))


def make_symlink (source ,target ):
    if os .path .islink (target ):
        os .unlink (target )

    os .symlink (source ,target )


def newer (source ,target ):
    """
    distutils.dep_utils.newer with zipfile support
    """
    try :
        return zipio .getmtime (source )>zipio .getmtime (target )
    except IOError :
        return True 


def find_version (fn ):
    """
    Try to find a __version__ assignment in a source file
    """
    return "0.0.0"
    import compiler 
    from compiler .ast import Assign ,AssName ,Const ,Module ,Stmt 

    ast =compiler .parseFile (fn )
    if not isinstance (ast ,Module ):
        raise ValueError ("expecting Module")
    statements =ast .getChildNodes ()
    if not (len (statements )==1 and isinstance (statements [0 ],Stmt )):
        raise ValueError ("expecting one Stmt")
    for node in statements [0 ].getChildNodes ():
        if not isinstance (node ,Assign ):
            continue 
        if not len (node .nodes )==1 :
            continue 
        assName =node .nodes [0 ]
        if not (
        isinstance (assName ,AssName )
        and isinstance (node .expr ,Const )
        and assName .flags =="OP_ASSIGN"
        and assName .name =="__version__"
        ):
            continue 
        return node .expr .value 
    else :
        raise ValueError ("Version not found")


def in_system_path (filename ):
    """
    Return True if the file is in a system path
    """
    return macholib .util .in_system_path (filename )


if sys .version_info [0 ]==2 :

    def fsencoding (s ,encoding =sys .getfilesystemencoding ()):
        return macholib .util .fsencoding (s ,encoding =encoding )

else :

    def fsencoding (s ,encoding =sys .getfilesystemencoding ()):
        return s 


def make_exec (path ):
    mask =os .umask (0 )
    os .umask (mask )
    os .chmod (path ,os .stat (path ).st_mode |(0o111 &~mask ))


def makedirs (path ):
    path =fsencoding (path )
    if not os .path .exists (path ):
        os .makedirs (path )


def mergecopy (src ,dest ):
    return macholib .util .mergecopy (src ,dest )


def mergetree (src ,dst ,condition =None ,copyfn =mergecopy ):
    """Recursively merge a directory tree using mergecopy()."""
    return macholib .util .mergetree (src ,dst ,condition =condition ,copyfn =copyfn )


def move (src ,dst ):
    return macholib .util .move (src ,dst )


def copy2 (src ,dst ):
    return macholib .util .copy2 (src ,dst )


def fancy_split (s ,sep =","):


    if s is None :
        return []
    if hasattr (s ,"split"):
        return [item .strip ()for item in s .split (sep )]
    return s 


class FileSet (object ):

    def __init__ (self ,iterable =None ):
        self ._dict ={}
        if iterable is not None :
            for arg in iterable :
                self .add (arg )

    def __repr__ (self ):
        return "<FileSet %s at %x>"%(self ._dict .values (),id (self ))

    def add (self ,fname ):
        self ._dict [fname .upper ()]=fname 

    def remove (self ,fname ):
        del self ._dict [fname .upper ()]

    def __contains__ (self ,fname ):
        return fname .upper ()in self ._dict .keys ()

    def __getitem__ (self ,index ):
        key =self ._dict .keys ()[index ]
        return self ._dict [key ]

    def __len__ (self ):
        return len (self ._dict )

    def copy (self ):
        res =FileSet ()
        res ._dict .update (self ._dict )
        return res 



if imp is None :
        LOADER =textwrap .dedent ("""\
                def __load():
                    import os, sys, importlib.machinery, importlib._bootstrap
                    ext = %r
                    for path in sys.path:
                        if not path.endswith('lib-dynload'):
                            continue
                        ext_path = os.path.join(path, ext)
                        if os.path.exists(ext_path):
                            loader = importlib.machinery.ExtensionFileLoader(__name__, ext_path)
                            spec = importlib.machinery.ModuleSpec(
                                name=__name__, loader=loader, origin=ext_path)
                            importlib._bootstrap._load(spec)
                            break
                    else:
                        raise ImportError(repr(ext) + " not found")
                __load()
                del __load
                """)

else :
        LOADER =textwrap .dedent ("""\
                def __load():
                    import os, sys, imp
                    ext = %r
                    for path in sys.path:
                        if not path.endswith('lib-dynload'):
                            continue
                        ext_path = os.path.join(path, ext)
                        if os.path.exists(ext_path):
                            mod = imp.load_dynamic(__name__, ext_path)
                            break
                    else:
                        raise ImportError(repr(ext) + " not found")
                __load()
                del __load
                """)


def make_loader (fn ):
    return LOADER %fn 


def byte_compile (
py_files ,optimize =0 ,force =0 ,target_dir =None ,verbose =1 ,dry_run =0 ,direct =None 
):

    if direct is None :
        direct =__debug__ and optimize ==0 



    if not direct :
        from distutils .util import execute ,spawn 
        from tempfile import NamedTemporaryFile 

        if verbose :
            print ("writing byte-compilation script")
        if not dry_run :
            kwargs ={"encoding":"utf-8"}if sys .version_info [0 ]==3 else {}
            with NamedTemporaryFile (
            suffix =".py",delete =False ,mode ="w",**kwargs 
            )as script :
                script_name =script .name 
                script .write (
                """
from py2app.util import byte_compile
from modulegraph.modulegraph import *
files = [
"""
                )

                for f in py_files :
                    script .write (repr (f )+",\n")
                script .write ("]\n")
                script .write (
                """
byte_compile(files, optimize=%r, force=%r,
             target_dir=%r,
             verbose=%r, dry_run=0,
             direct=1)
"""
                %(optimize ,force ,target_dir ,verbose )
                )




        import py2app 

        pp =os .path .dirname (os .path .dirname (py2app .__file__ ))
        if "PYTHONPATH"in os .environ :
            pp ="%s:%s"%(pp ,os .environ ["PYTHONPATH"])

        cmd =["/usr/bin/env","PYTHONPATH=%s"%(pp ,),sys .executable ,script_name ]

        if optimize ==1 :
            cmd .insert (3 ,"-O")
        elif optimize ==2 :
            cmd .insert (3 ,"-OO")
        spawn (cmd ,verbose =verbose ,dry_run =dry_run )
        execute (
        os .remove ,
        (script_name ,),
        "removing %s"%script_name ,
        verbose =verbose ,
        dry_run =dry_run ,
        )

    else :
        from distutils .dir_util import mkpath 
        from py_compile import compile 

        for mod in py_files :



            if mod .filename ==mod .identifier :
                cfile =os .path .basename (mod .filename )
                dfile =cfile +(__debug__ and "c"or "o")
            else :
                cfile =mod .identifier .replace (".",os .sep )

                if sys .version_info [:2 ]<=(3 ,4 ):
                    if mod .packagepath :
                        dfile =(
                        cfile +os .sep +"__init__.py"+(__debug__ and "c"or "o")
                        )
                    else :
                        dfile =cfile +".py"+(__debug__ and "c"or "o")
                else :
                    if mod .packagepath :
                        dfile =cfile +os .sep +"__init__.pyc"
                    else :
                        dfile =cfile +".pyc"
            if target_dir :
                cfile =os .path .join (target_dir ,dfile )

            if force or newer (mod .filename ,cfile ):
                if verbose :
                    print ("byte-compiling %s to %s"%(mod .filename ,dfile ))

                if not dry_run :
                    mkpath (os .path .dirname (cfile ))
                    suffix =os .path .splitext (mod .filename )[1 ]

                    if suffix in (".py",".pyw"):
                        fn =cfile +".py"

                        with zipio .open (mod .filename ,"rb")as fp_in :
                            with open (fn ,"wb")as fp_out :
                                fp_out .write (fp_in .read ())

                        compile (fn ,cfile ,dfile )
                        os .unlink (fn )

                    elif suffix in PY_SUFFIXES :



                        copy_file (mod .filename ,cfile ,preserve_times =True )

                    else :
                        raise RuntimeError ("Don't know how to handle %r"%mod .filename )
            else :
                if verbose :
                    print (
                    "skipping byte-compilation of %s to %s"%(mod .filename ,dfile )
                    )


SCMDIRS =["CVS",".svn",".hg",".git"]


def skipscm (ofn ):
    ofn =fsencoding (ofn )
    fn =os .path .basename (ofn )
    if fn in SCMDIRS :
        return False 
    return True 


def skipfunc (junk =(),junk_exts =(),chain =()):
    junk =set (junk )
    junk_exts =set (junk_exts )
    chain =tuple (chain )

    def _skipfunc (fn ):
        if os .path .basename (fn )in junk :
            return False 
        elif os .path .splitext (fn )[1 ]in junk_exts :
            return False 
        for func in chain :
            if not func (fn ):
                return False 
        else :
            return True 

    return _skipfunc 


JUNK =[".DS_Store",".gdb_history","build","dist"]+SCMDIRS 
JUNK_EXTS =[".pbxuser",".pyc",".pyo",".swp"]
skipjunk =skipfunc (JUNK ,JUNK_EXTS )


def get_magic (platform =sys .platform ):
    if platform =="darwin":
        import struct 

        import macholib .mach_o 

        return [
        struct .pack ("!L",macholib .mach_o .MH_MAGIC ),
        struct .pack ("!L",macholib .mach_o .MH_CIGAM ),
        struct .pack ("!L",macholib .mach_o .MH_MAGIC_64 ),
        struct .pack ("!L",macholib .mach_o .MH_CIGAM_64 ),
        struct .pack ("!L",macholib .mach_o .FAT_MAGIC ),
        ]
    elif platform =="linux2":
        return ["\x7fELF"]
    elif platform =="win32":
        return ["MZ"]
    return None 


def iter_platform_files (path ,is_platform_file =macholib .util .is_platform_file ):
    """
    Iterate over all of the platform files in a directory
    """
    for root ,_dirs ,files in os .walk (path ):
        for fn in files :
            fn =os .path .join (root ,fn )
            if is_platform_file (fn ):
                yield fn 


def strip_files (files ,dry_run =0 ,verbose =0 ):
    """
    Strip the given set of files
    """
    if dry_run :
        return 

    with reset_blocking_status ():
        return macholib .util .strip_files (files )


def copy_tree (
src ,
dst ,
preserve_mode =1 ,
preserve_times =1 ,
preserve_symlinks =0 ,
update =0 ,
verbose =0 ,
dry_run =0 ,
condition =None ,
):

    """
    Copy an entire directory tree 'src' to a new location 'dst'.  Both
    'src' and 'dst' must be directory names.  If 'src' is not a
    directory, raise DistutilsFileError.  If 'dst' does not exist, it is
    created with 'mkpath()'.  The end result of the copy is that every
    file in 'src' is copied to 'dst', and directories under 'src' are
    recursively copied to 'dst'.  Return the list of files that were
    copied or might have been copied, using their output name.  The
    return value is unaffected by 'update' or 'dry_run': it is simply
    the list of all files under 'src', with the names changed to be
    under 'dst'.

    'preserve_mode' and 'preserve_times' are the same as for
    'copy_file'; note that they only apply to regular files, not to
    directories.  If 'preserve_symlinks' is true, symlinks will be
    copied as symlinks (on platforms that support them!); otherwise
    (the default), the destination of the symlink will be copied.
    'update' and 'verbose' are the same as for 'copy_file'.
    """
    assert isinstance (src ,(str ,unicode )),repr (src )
    assert isinstance (dst ,(str ,unicode )),repr (dst )

    from distutils import log 
    from distutils .dep_util import newer 
    from distutils .dir_util import mkpath 
    from distutils .errors import DistutilsFileError 

    src =fsencoding (src )
    dst =fsencoding (dst )

    if condition is None :
        condition =skipscm 

    if not dry_run and not zipio .isdir (src ):
        raise DistutilsFileError ("cannot copy tree '%s': not a directory"%src )
    try :
        names =zipio .listdir (src )
    except os .error as exc :
        (errno ,errstr )=exc .args 
        if dry_run :
            names =[]
        else :
            raise DistutilsFileError ("error listing files in '%s': %s"%(src ,errstr ))

    if not dry_run :
        mkpath (dst )

    outputs =[]

    for n in names :
        src_name =os .path .join (src ,n )
        dst_name =os .path .join (dst ,n )
        if (condition is not None )and (not condition (src_name )):
            continue 



        if os .path .islink (src_name )and not os .path .exists (
        os .path .join (src ,os .readlink (src_name ))
        ):
            continue 

        if preserve_symlinks and zipio .islink (src_name ):
            link_dest =zipio .readlink (src_name )
            log .info ("linking %s -> %s",dst_name ,link_dest )
            if not dry_run :
                if update and not newer (src ,dst_name ):
                    pass 
                else :
                    make_symlink (link_dest ,dst_name )
            outputs .append (dst_name )

        elif zipio .isdir (src_name )and not os .path .isfile (src_name ):


            outputs .extend (
            copy_tree (
            src_name ,
            dst_name ,
            preserve_mode ,
            preserve_times ,
            preserve_symlinks ,
            update ,
            dry_run =dry_run ,
            condition =condition ,
            )
            )
        else :
            copy_file (
            src_name ,
            dst_name ,
            preserve_mode ,
            preserve_times ,
            update ,
            dry_run =dry_run ,
            )
            outputs .append (dst_name )

    return outputs 


def walk_files (path ):
    for _root ,_dirs ,files in os .walk (path ):
        for f in files :
            yield f 


def find_app (app ):
    dpath =os .path .realpath (app )
    if os .path .exists (dpath ):
        return dpath 
    if os .path .isabs (app ):
        return None 
    for path in os .environ .get ("PATH","").split (":"):
        dpath =os .path .realpath (os .path .join (path ,app ))
        if os .path .exists (dpath ):
            return dpath 
    return None 


def check_output (command_line ):
    p =subprocess .Popen (command_line ,stdout =subprocess .PIPE )
    stdout ,_ =p .communicate ()
    xit =p .wait ()
    if xit !=0 :
        raise subprocess .CalledProcessError (xit ,command_line )

    return stdout 


_tools ={}


def _get_tool (toolname ):
    if toolname not in _tools :
        if os .path .exists ("/usr/bin/xcrun"):
            try :
                _tools [toolname ]=check_output (["/usr/bin/xcrun","-find",toolname ])[
                :-1 
                ]
            except subprocess .CalledProcessError :
                raise IOError ("Tool %r not found"%(toolname ,))

        else :

            if toolname =="momc":
                choices =[
                (
                "/Library/Application Support/Apple/"
                "Developer Tools/Plug-ins/XDCoreDataModel.xdplugin/"
                "Contents/Resources/momc"
                ),
                (
                "/Developer/Library/Xcode/Plug-ins/"
                "XDCoreDataModel.xdplugin/Contents/Resources/momc"
                ),
                "/Developer/usr/bin/momc",
                ]
            elif toolname =="mapc":
                choices =[
                (
                "/Developer/Library/Xcode/Plug-ins/"
                "XDMappingModel.xdplugin/"
                "Contents/Resources/mapc",
                ),
                "/Developer/usr/bin/mapc",
                ]
            else :
                raise IOError ("Tool %r not found"%(toolname ,))

            for fn in choices :
                if os .path .exists (fn ):
                    _tools [toolname ]=fn 
                    break 
            else :
                raise IOError ("Tool %r not found"%(toolname ,))
    return _tools [toolname ]


def momc (src ,dst ):
    with reset_blocking_status ():
        subprocess .check_call ([_get_tool ("momc"),src ,dst ])


def mapc (src ,dst ):
    with reset_blocking_status ():
        subprocess .check_call ([_get_tool ("mapc"),src ,dst ])


def _macho_find (path ):
    for basename ,_dirs ,files in os .walk (path ):
        for fn in files :
            path =os .path .join (basename ,fn )
            if is_platform_file (path ):
                yield path 


def _dosign (*path ):
    with reset_blocking_status ():
        subprocess .check_call (
        (
        "codesign",
        "-s",
        "-",
        "--preserve-metadata=identifier,entitlements,flags,runtime",
        "-f",
        "-vvvv",
        )
        +path ,
        )


def codesign_adhoc (bundle ):
    """
    (Re)sign a bundle

    Signing should be done "depth-first", sign
    libraries before signing the libraries/executables
    linking to them.

    The current implementation is a crude hack,
    but is better than nothing. Signing properly requires
    performing a topological sort using dependencies.

    "codesign" will resign the entire bundle, but only
    if partial signatures are valid.
    """






    platfiles =list (_macho_find (bundle ))
    print ("sign",platfiles )
    while platfiles :
        for file in platfiles :
            failed =[]
            try :
                _dosign (file )
            except subprocess .CalledProcessError :
                failed .append (file )
        if failed ==platfiles :
            raise RuntimeError ("Cannot sign bundle %r"%(bundle ,))
        platfiles =failed 

    for _ in range (5 ):
        try :
            _dosign (bundle )
            break 
        except subprocess .CalledProcessError :
            time .sleep (1 )
            continue 


@contextlib .contextmanager 
def reset_blocking_status ():
    """
    Contextmanager that resets the non-blocking status of
    the std* streams as necessary. Used with all calls of
    xcode tools, mostly because ibtool tends to set the
    std* streams to non-blocking.
    """
    orig_nonblocking =[
    fcntl .fcntl (fd ,fcntl .F_GETFL )&os .O_NONBLOCK for fd in (0 ,1 ,2 )
    ]

    try :
        yield 

    finally :
        for fd ,is_nonblocking in zip ((0 ,1 ,2 ),orig_nonblocking ):
            cur =fcntl .fcntl (fd ,fcntl .F_GETFL )
            if is_nonblocking :
                reset =cur |os .O_NONBLOCK 
            else :
                reset =cur &~os .O_NONBLOCK 

            if cur !=reset :
                print ("Resetting blocking status of %s"%(fd ,))
                fcntl .fcntl (fd ,fcntl .F_SETFL ,reset )



"""
Utilities for reading and writing Mach-O headers
"""
from __future__ import print_function 

import os 
import struct 
import sys 

from macholib .util import fileview 

from .mach_o import (
FAT_MAGIC ,
FAT_MAGIC_64 ,
LC_DYSYMTAB ,
LC_ID_DYLIB ,
LC_LOAD_DYLIB ,
LC_LOAD_UPWARD_DYLIB ,
LC_LOAD_WEAK_DYLIB ,
LC_PREBOUND_DYLIB ,
LC_REEXPORT_DYLIB ,
LC_REGISTRY ,
LC_SEGMENT ,
LC_SEGMENT_64 ,
LC_SYMTAB ,
MH_CIGAM ,
MH_CIGAM_64 ,
MH_FILETYPE_SHORTNAMES ,
MH_MAGIC ,
MH_MAGIC_64 ,
S_ZEROFILL ,
fat_arch ,
fat_arch64 ,
fat_header ,
load_command ,
mach_header ,
mach_header_64 ,
section ,
section_64 ,
)
from .ptypes import sizeof 

try :
    from macholib .compat import bytes 
except ImportError :
    pass 

try :
    unicode 
except NameError :
    unicode =str 

if sys .version_info [0 ]==2 :
    range =xrange 

__all__ =["MachO"]

_RELOCATABLE ={

LC_LOAD_DYLIB ,
LC_LOAD_UPWARD_DYLIB ,
LC_LOAD_WEAK_DYLIB ,
LC_PREBOUND_DYLIB ,
LC_REEXPORT_DYLIB ,
}

_RELOCATABLE_NAMES ={
LC_LOAD_DYLIB :"load_dylib",
LC_LOAD_UPWARD_DYLIB :"load_upward_dylib",
LC_LOAD_WEAK_DYLIB :"load_weak_dylib",
LC_PREBOUND_DYLIB :"prebound_dylib",
LC_REEXPORT_DYLIB :"reexport_dylib",
}


def _shouldRelocateCommand (cmd ):
    """
    Should this command id be investigated for relocation?
    """
    return cmd in _RELOCATABLE 


def lc_str_value (offset ,cmd_info ):
    """
    Fetch the actual value of a field of type "lc_str"
    """
    cmd_load ,cmd_cmd ,cmd_data =cmd_info 

    offset -=sizeof (cmd_load )+sizeof (cmd_cmd )
    return cmd_data [offset :].strip (b"\x00")


class MachO (object ):
    """
    Provides reading/writing the Mach-O header of a specific existing file.

    If allow_unknown_load_commands is True, allows unknown load commands.
    Otherwise, raises ValueError if the file contains an unknown load command.
    """










    def __init__ (self ,filename ,allow_unknown_load_commands =False ):

        self .graphident =filename 
        self .filename =filename 
        self .loader_path =os .path .dirname (filename )


        self .fat =None 
        self .headers =[]
        self .allow_unknown_load_commands =allow_unknown_load_commands 
        with open (filename ,"rb")as fp :
            self .load (fp )

    def __repr__ (self ):
        return "<MachO filename=%r>"%(self .filename ,)

    def load (self ,fh ):
        assert fh .tell ()==0 
        header =struct .unpack (">I",fh .read (4 ))[0 ]
        fh .seek (0 )
        if header in (FAT_MAGIC ,FAT_MAGIC_64 ):
            self .load_fat (fh )
        else :
            fh .seek (0 ,2 )
            size =fh .tell ()
            fh .seek (0 )
            self .load_header (fh ,0 ,size )

    def load_fat (self ,fh ):
        self .fat =fat_header .from_fileobj (fh )
        if self .fat .magic ==FAT_MAGIC :
            archs =[fat_arch .from_fileobj (fh )for i in range (self .fat .nfat_arch )]
        elif self .fat .magic ==FAT_MAGIC_64 :
            archs =[fat_arch64 .from_fileobj (fh )for i in range (self .fat .nfat_arch )]
        else :
            raise ValueError ("Unknown fat header magic: %r"%(self .fat .magic ))

        for arch in archs :
            self .load_header (fh ,arch .offset ,arch .size )

    def rewriteLoadCommands (self ,*args ,**kw ):
        changed =False 
        for header in self .headers :
            if header .rewriteLoadCommands (*args ,**kw ):
                changed =True 
        return changed 

    def load_header (self ,fh ,offset ,size ):
        fh .seek (offset )
        header =struct .unpack (">I",fh .read (4 ))[0 ]
        fh .seek (offset )
        if header ==MH_MAGIC :
            magic ,hdr ,endian =MH_MAGIC ,mach_header ,">"
        elif header ==MH_CIGAM :
            magic ,hdr ,endian =MH_CIGAM ,mach_header ,"<"
        elif header ==MH_MAGIC_64 :
            magic ,hdr ,endian =MH_MAGIC_64 ,mach_header_64 ,">"
        elif header ==MH_CIGAM_64 :
            magic ,hdr ,endian =MH_CIGAM_64 ,mach_header_64 ,"<"
        else :
            raise ValueError ("Unknown Mach-O header: 0x%08x in %r"%(header ,fh ))
        hdr =MachOHeader (
        self ,fh ,offset ,size ,magic ,hdr ,endian ,self .allow_unknown_load_commands 
        )
        self .headers .append (hdr )

    def write (self ,f ):
        for header in self .headers :
            header .write (f )


class MachOHeader (object ):
    """
    Provides reading/writing the Mach-O header of a specific existing file.

    If allow_unknown_load_commands is True, allows unknown load commands.
    Otherwise, raises ValueError if the file contains an unknown load command.
    """










    def __init__ (
    self ,
    parent ,
    fh ,
    offset ,
    size ,
    magic ,
    hdr ,
    endian ,
    allow_unknown_load_commands =False ,
    ):
        self .MH_MAGIC =magic 
        self .mach_header =hdr 


        self .parent =parent 
        self .offset =offset 
        self .size =size 

        self .endian =endian 
        self .header =None 
        self .commands =None 
        self .id_cmd =None 
        self .sizediff =None 
        self .total_size =None 
        self .low_offset =None 
        self .filetype =None 
        self .headers =[]

        self .allow_unknown_load_commands =allow_unknown_load_commands 

        self .load (fh )

    def __repr__ (self ):
        return "<%s filename=%r offset=%d size=%d endian=%r>"%(
        type (self ).__name__ ,
        self .parent .filename ,
        self .offset ,
        self .size ,
        self .endian ,
        )

    def load (self ,fh ):
        fh =fileview (fh ,self .offset ,self .size )
        fh .seek (0 )

        self .sizediff =0 
        kw ={"_endian_":self .endian }
        header =self .mach_header .from_fileobj (fh ,**kw )
        self .header =header 




        cmd =self .commands =[]

        self .filetype =self .get_filetype_shortname (header .filetype )

        read_bytes =0 
        low_offset =sys .maxsize 
        for i in range (header .ncmds ):

            cmd_load =load_command .from_fileobj (fh ,**kw )


            klass =LC_REGISTRY .get (cmd_load .cmd ,None )
            if klass is None :
                if not self .allow_unknown_load_commands :
                    raise ValueError ("Unknown load command: %d"%(cmd_load .cmd ,))


                data_size =cmd_load .cmdsize -sizeof (load_command )
                cmd_data =fh .read (data_size )
                cmd .append ((cmd_load ,cmd_load ,cmd_data ))
                read_bytes +=cmd_load .cmdsize 
                continue 

            cmd_cmd =klass .from_fileobj (fh ,**kw )

            if cmd_load .cmd ==LC_ID_DYLIB :

                if self .id_cmd is not None :
                    raise ValueError ("This dylib already has an id")
                self .id_cmd =i 

            if cmd_load .cmd in (LC_SEGMENT ,LC_SEGMENT_64 ):

                segs =[]

                if cmd_load .cmd ==LC_SEGMENT :
                    section_cls =section 
                else :
                    section_cls =section_64 

                expected_size =(
                sizeof (klass )
                +sizeof (load_command )
                +(sizeof (section_cls )*cmd_cmd .nsects )
                )
                if cmd_load .cmdsize !=expected_size :
                    raise ValueError ("Segment size mismatch")


                if cmd_cmd .nsects ==0 :
                    if cmd_cmd .filesize !=0 :
                        low_offset =min (low_offset ,cmd_cmd .fileoff )
                else :

                    for _j in range (cmd_cmd .nsects ):

                        seg =section_cls .from_fileobj (fh ,**kw )


                        not_zerofill =(seg .flags &S_ZEROFILL )!=S_ZEROFILL 
                        if seg .offset >0 and seg .size >0 and not_zerofill :
                            low_offset =min (low_offset ,seg .offset )
                        if not_zerofill :
                            c =fh .tell ()
                            fh .seek (seg .offset )
                            sd =fh .read (seg .size )
                            seg .add_section_data (sd )
                            fh .seek (c )
                        segs .append (seg )

                cmd_data =segs 













            else :

                data_size =cmd_load .cmdsize -sizeof (klass )-sizeof (load_command )
                cmd_data =fh .read (data_size )
            cmd .append ((cmd_load ,cmd_cmd ,cmd_data ))
            read_bytes +=cmd_load .cmdsize 


        if read_bytes !=header .sizeofcmds :
            raise ValueError (
            "Read %d bytes, header reports %d bytes"
            %(read_bytes ,header .sizeofcmds )
            )
        self .total_size =sizeof (self .mach_header )+read_bytes 
        self .low_offset =low_offset 

    def walkRelocatables (self ,shouldRelocateCommand =_shouldRelocateCommand ):
        """
        for all relocatable commands
        yield (command_index, command_name, filename)
        """
        for idx ,(lc ,cmd ,data )in enumerate (self .commands ):
            if shouldRelocateCommand (lc .cmd ):
                name =_RELOCATABLE_NAMES [lc .cmd ]
                ofs =cmd .name -sizeof (lc .__class__ )-sizeof (cmd .__class__ )
                yield idx ,name ,data [
                ofs :data .find (b"\x00",ofs )
                ].decode (sys .getfilesystemencoding ())

    def rewriteInstallNameCommand (self ,loadcmd ):
        """Rewrite the load command of this dylib"""
        if self .id_cmd is not None :
            self .rewriteDataForCommand (self .id_cmd ,loadcmd )
            return True 
        return False 

    def changedHeaderSizeBy (self ,bytes ):
        self .sizediff +=bytes 
        if (self .total_size +self .sizediff )>self .low_offset :
            print (
            "WARNING: Mach-O header in %r may be too large to relocate"
            %(self .parent .filename ,)
            )

    def rewriteLoadCommands (self ,changefunc ):
        """
        Rewrite the load commands based upon a change dictionary
        """
        data =changefunc (self .parent .filename )
        changed =False 
        if data is not None :
            if self .rewriteInstallNameCommand (data .encode (sys .getfilesystemencoding ())):
                changed =True 
        for idx ,_name ,filename in self .walkRelocatables ():
            data =changefunc (filename )
            if data is not None :
                if self .rewriteDataForCommand (
                idx ,data .encode (sys .getfilesystemencoding ())
                ):
                    changed =True 
        return changed 

    def rewriteDataForCommand (self ,idx ,data ):
        lc ,cmd ,old_data =self .commands [idx ]
        hdrsize =sizeof (lc .__class__ )+sizeof (cmd .__class__ )
        align =struct .calcsize ("Q")
        data =data +(b"\x00"*(align -(len (data )%align )))
        newsize =hdrsize +len (data )
        self .commands [idx ]=(lc ,cmd ,data )
        self .changedHeaderSizeBy (newsize -lc .cmdsize )
        lc .cmdsize ,cmd .name =newsize ,hdrsize 
        return True 

    def synchronize_size (self ):
        if (self .total_size +self .sizediff )>self .low_offset :
            raise ValueError (
            (
            "New Mach-O header is too large to relocate in %r "
            "(new size=%r, max size=%r, delta=%r)"
            )
            %(
            self .parent .filename ,
            self .total_size +self .sizediff ,
            self .low_offset ,
            self .sizediff ,
            )
            )
        self .header .sizeofcmds +=self .sizediff 
        self .total_size =sizeof (self .mach_header )+self .header .sizeofcmds 
        self .sizediff =0 

    def write (self ,fileobj ):
        fileobj =fileview (fileobj ,self .offset ,self .size )
        fileobj .seek (0 )


        self .synchronize_size ()

        self .header .to_fileobj (fileobj )
        for lc ,cmd ,data in self .commands :
            lc .to_fileobj (fileobj )
            cmd .to_fileobj (fileobj )

            if sys .version_info [0 ]==2 :
                if isinstance (data ,unicode ):
                    fileobj .write (data .encode (sys .getfilesystemencoding ()))

                elif isinstance (data ,(bytes ,str )):
                    fileobj .write (data )
                else :

                    for obj in data :
                        obj .to_fileobj (fileobj )
            else :
                if isinstance (data ,str ):
                    fileobj .write (data .encode (sys .getfilesystemencoding ()))

                elif isinstance (data ,bytes ):
                    fileobj .write (data )

                else :

                    for obj in data :
                        obj .to_fileobj (fileobj )



        fileobj .write (b"\x00"*(self .low_offset -fileobj .tell ()))

    def getSymbolTableCommand (self ):
        for lc ,cmd ,_data in self .commands :
            if lc .cmd ==LC_SYMTAB :
                return cmd 
        return None 

    def getDynamicSymbolTableCommand (self ):
        for lc ,cmd ,_data in self .commands :
            if lc .cmd ==LC_DYSYMTAB :
                return cmd 
        return None 

    def get_filetype_shortname (self ,filetype ):
        if filetype in MH_FILETYPE_SHORTNAMES :
            return MH_FILETYPE_SHORTNAMES [filetype ]
        else :
            return "unknown"


def main (fn ):
    m =MachO (fn )
    seen =set ()
    for header in m .headers :
        for _idx ,name ,other in header .walkRelocatables ():
            if other not in seen :
                seen .add (other )
                print ("\t"+name +": "+other )


if __name__ =="__main__":
    import sys 

    files =sys .argv [1 :]or ["/bin/ls"]
    for fn in files :
        print (fn )
        main (fn )
